/*
 * Copyright (C) 2002  Olaf Beck <olaf_sc@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "tydemux.h"
#include "global.h"

void free_audio_pes(audio_pes_t * audio_pes) {
	free(audio_pes->pes_header);
	free(audio_pes->pes_data);
	free(audio_pes);
}

void free_audio_pess(audio_pes_t * audio_pes) {

	audio_pes_t * free_pes;

	if(audio_pes != NULL) {
		while(audio_pes->next != NULL){
			free_pes = audio_pes;
			audio_pes = audio_pes->next;
			free_audio_pes(free_pes);
		}

		free_audio_pes(audio_pes);
	}
}

/* Old debug function - will be removed */
audio_pes_t * owrite_audio_pes(audio_pes_t * audio_pes) {

	audio_pes_t * next_pes;
#ifdef DEBUG
	int k;
	uint8_t comp[]={0x00, 0x00, 0x01, 0xc0, 0x01, 0xe5, 0x21};
	uint8_t comp1[]={0xff, 0xfd, 0x94, 0x40};
#endif
	nr_audio_pes--;

	if( debug >= 6) {
		printf("Write Audio: writing pes with time: %llu  size: %d\n",audio_pes->time,audio_pes->size);
	}

	if(audio_pes->next != NULL) {
		/* Writing :) */
		//write(audio_file_nr, audio_pes->pes_header, 14);
		if(write_file) {
			write(audio_file_nr, audio_pes->pes_data, audio_pes->size);
		}
#ifdef DEBUG
		/* Compile with -DDEBUG if you want to check mpeg-II audio errors */
		if (memcmp(audio_pes->pes_header, comp, 7) != 0) {
			printf("Header:");
			for(k=0; k< 11; k++){
				printf(" %02x", audio_pes->pes_header[k]);
			}
			printf("\n");
		}
		if (memcmp(audio_pes->pes_data, comp1, 4) != 0) {
			printf("Body:");
			for(k=0; k< 16; k++){
				printf(" %02x", audio_pes->pes_data[k]);
			}
			printf("\n");
		}
#endif
		next_pes=audio_pes->next;
		next_pes->previous = NULL;
		free_audio_pes(audio_pes);
		return(next_pes);
	} else {
		/* last pes to write */
		if(write_file){
			write(audio_file_nr, audio_pes->pes_data, audio_pes->size);
		}
		free(audio_pes);
		return(0);
	}
}

#ifdef A52
int a52_info(uint8_t * pt1) {

	int length;
	static int sample_rate;
	static int flags;
	int bit_rate;


	length = a52_syncinfo (pt1, &flags, &sample_rate, &bit_rate);
	if (!length) {
		if(debug >= 1) {
			printf("A52 Info: Invalid AC3 record\n");
		}
		return(-1);
	} else {
		if(debug >= 2) {
			printf("A52 Info: Samplerate: %d, Bitrate %d\n",sample_rate,bit_rate);
			printf("A52 Info: Channles %01x\n", flags&A52_CHANNEL_MASK);
			printf("A52 Info: LFE %01x\n", (flags&A52_LFE) >> 4);
		}
		return(flags&A52_CHANNEL_MASK);
	}

}

audio_pes_t * transcode_ac3(audio_pes_t * audio_pes, int flags) {

	/* interation */
	int i;

	/* Channels */
	int channels;

	/* Decoding */

	AVCodec *decode;
	AVCodecContext decode_context, *d = &decode_context;
	int out_size, len;
	uint8_t * decode_buffer;
	uint8_t * inbuf_ptr;

	/* Encoding */

	AVCodec *codec;
	AVCodecContext codec_context, *c = &codec_context;
	int  size, tmp_encoded_buffer_size;
	uint8_t * tmp_encoded_buffer;
	uint8_t * encoded_buffer;

	/* Pes audio data */
	uint8_t * old_pes_data;
	//printf("Transcode:\n");

	/* Now lets decode the audio_pes->pes_data buffer */

	/* Find the ac3 audio decoder */
	decode = avcodec_find_decoder(CODEC_ID_AC3);
	if (!decode) {
		if(debug >= 1) {
			printf("Transcode: Avcodec find error\n");
		}
		return(audio_pes);
	}

	/* Put default values in d*/
	memset(d, 0, sizeof(*d));

	/* Open ac3 codec */
	if (avcodec_open(d, decode) < 0) {
		if(debug >= 1) {
			printf("Transcode: Avcodec load error\n");
		}
		avcodec_close(d);
		return(audio_pes);
	}

	/* Allocate our decoded audio buffer */
	decode_buffer = malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);
	memset(decode_buffer, 0,AVCODEC_MAX_AUDIO_FRAME_SIZE);

	inbuf_ptr = audio_pes->pes_data;

	/* Decode it */
	len = avcodec_decode_audio(d, (short *)decode_buffer, &out_size,
		inbuf_ptr, audio_pes->size);


	if (len < 0) {
		if(debug >= 1) {
			printf("Transcode: Avcodec decoding error\n");
		}
		free(decode_buffer);
		avcodec_close(d);
		return(audio_pes);
	}


	channels = d->channels;
	//printf("Transcode: channels  %d\n", channels);

	if (out_size > 0) {
		//printf("Transcode: Writing\n");
		fwrite(decode_buffer, 1, out_size, dbgfile);
	}
	//} else {
	//	printf("Transcode: outsize %i %i\n", out_size, len);
		//for(i=0; i < 16 ; i++){
		//	printf(" %02x", decode_buffer[i]);
		//}
		//printf("\n");
	//	free(decode_buffer);
	//	avcodec_close(d);
	//	return(audio_pes);
	//}

	/* Close the codec FIXME need to check if this is really nessesary */
	avcodec_close(d);

	/* Now lets encode the decode_buffer */

	/* Find the AC3 encoder FIXME need to check if this is redundant*/
	//codec = avcodec_find_encoder(CODEC_ID_AC3);
	codec = avcodec_find_encoder(CODEC_ID_MP2);
	if (!codec) {
		if(debug >= 1) {
			printf("Transcode: Avcodec find (in encode) error\n");
		}
		free(decode_buffer);
		return(audio_pes);
	}

	/* Put default values in c*/
	memset(c, 0, sizeof(*c));

	/* Set encoding parameters to the SVCD specs */
	c->bit_rate = 224000;
	c->sample_rate = 44100;
	/* Would like to set this to
	c->channels = flags
	but it looks like ffmpeg can't encode 5.1 Dolby digital */
	c->channels = channels;

	/* Open the AC3 codec again */
	if (avcodec_open(c, codec) < 0) {
		if(debug >= 1) {
			printf("Transcode: Avcodec load (in encode) error\n");
		}
		free(decode_buffer);
		avcodec_close(c);
		return(audio_pes);
	}

	/* Allocate our encoded buffer (a bit big but what the heck */
	tmp_encoded_buffer_size = 10000;
	tmp_encoded_buffer = malloc(tmp_encoded_buffer_size);

	/* Encode our decoded buffer */
	size = avcodec_encode_audio(c, tmp_encoded_buffer, tmp_encoded_buffer_size, (short *)decode_buffer);
	avcodec_close(c);

	if(size > 0) {
		/* Swap the audio_pes->pes_data with our encoded buffer */
		encoded_buffer = (uint8_t *)malloc(sizeof(uint8_t) * size);
		encoded_buffer = memcpy(encoded_buffer, tmp_encoded_buffer, size);
		if(encoded_buffer != NULL){
			free(audio_pes->pes_data);
			audio_pes->pes_data = encoded_buffer;
			audio_pes->size = size;
			free(tmp_encoded_buffer);
			free(decode_buffer);
			return(audio_pes);
		} else {
			free(tmp_encoded_buffer);
			free(decode_buffer);
			return(audio_pes);
		}
	} else {
		free(tmp_encoded_buffer);
		free(decode_buffer);
		return(audio_pes);
	}

}


#endif

audio_pes_t * write_audio_pes(audio_pes_t * audio_pes) {

	audio_pes_t * next_pes;
#ifdef A52
	uint8_t * pt1;
	int flags;

#endif
	nr_audio_pes--;

	if( debug >= 3) {
		printf("Write Audio: writing pes with time: %llu  size: %d\n",audio_pes->time,audio_pes->size);
	}

	if(audio_pes->next != NULL) {
		/* Writing :) */
		if(!askipp) {
		//write(audio_file_nr, audio_pes->pes_header, 14);

			if(write_file) {
#ifdef A52
				pt1 = audio_pes->pes_data;
				if(a52_okay) {
					if((flags = a52_info(pt1)) == -1) {
						if(debug >=1) {
							printf("Write Audio: A52 Flag error\n");
						}
					} else {
						audio_pes = transcode_ac3(audio_pes, flags);
						//pt1 = audio_pes->pes_data;
						//a52_info(pt1);
					}
				}
#endif
				write(audio_file_nr, audio_pes->pes_data, audio_pes->size);
			}


			next_pes=audio_pes->next;
			next_pes->previous = NULL;
			free_audio_pes(audio_pes);
			return(next_pes);
		} else {
			if(audio_pes->time <= video_seq_start && audio_pes->next->time >= video_seq_start ) {
				/* We got a match
				Lets see which one is the closest*/
				if(debug >= 3) {
					printf("Write Audio: Delay match - time %llu - next time %llu - video %llu\n", audio_pes->time,audio_pes->next->time,video_seq_start);
				}

				if((video_seq_start - audio_pes->time) <= (audio_pes->next->time - video_seq_start)) {
					/* Okay we should write this one and set the delay to XXX ticks
					early start of video stream i.e positive value*/

					if(write_file) {
#ifdef A52
						pt1 = audio_pes->pes_data;
						if(a52_okay) {
							if((flags = a52_info(pt1)) == -1) {
								if(debug >=1) {
									printf("Write Audio: A52 Flag error\n");
								}
							} else {
								audio_pes = transcode_ac3(audio_pes, flags);
							}
						}

#endif
						write(audio_file_nr, audio_pes->pes_data, audio_pes->size);
					}
					es_delay = video_seq_start - audio_pes->time;
					askipp = 0;
					next_pes=audio_pes->next;
					next_pes->previous = NULL;
					free_audio_pes(audio_pes);

					if(debug >= 3) {
						printf("Write Audio: Delay match 1 %lld other %llu \n", es_delay, audio_pes->next->time - video_seq_start);
					}
					return(next_pes);
				} else {
					/* Jump to next pes and write it set the delay to XXX ticks
					late start of video stream i.e negative value*/
					next_pes=audio_pes->next;
					next_pes->previous = NULL;
					free_audio_pes(audio_pes);
					audio_pes = next_pes;

					if(write_file) {
#ifdef A52
						pt1 = audio_pes->pes_data;
						if(a52_okay) {
							if((flags = a52_info(pt1)) == -1) {
								if(debug >=1) {
									printf("Write Audio: A52 Flag error\n");
								}
							} else {
								audio_pes = transcode_ac3(audio_pes, flags);
							}
						}

#endif
						write(audio_file_nr, audio_pes->pes_data, audio_pes->size);
					}

					es_delay = video_seq_start - audio_pes->time;
					askipp = 0;

					/* FIXME we should check the next pes isn't NULL */
					next_pes=audio_pes->next;
					next_pes->previous = NULL;
					free_audio_pes(audio_pes);

					if(debug >= 3) {
						printf("Write Audio: Delay match 2 %lld - other %llu\n", es_delay, video_seq_start - audio_pes->time);
					}
					return(next_pes);
				}

			} else {
				/* bummer not this one either :) */
				next_pes=audio_pes->next;
				next_pes->previous = NULL;
				free_audio_pes(audio_pes);
				return(next_pes);
			}
		}
	} else {
		/* last pes to write */
		if(write_file){
			write(audio_file_nr, audio_pes->pes_data, audio_pes->size);
		}
		free(audio_pes);
		return(0);
	}
}


audio_pes_t * insert_audio_pes(audio_pes_t * audio_pes){

	audio_pes_t * tmp_audio_pes;


	/* Increase the audio pes counter */
	nr_audio_pes++;

	tmp_audio_pes = m_audio_pes;

	while (tmp_audio_pes->next) {
		tmp_audio_pes = tmp_audio_pes->next;
	}
	tmp_audio_pes->next = audio_pes;
	audio_pes->previous = tmp_audio_pes;
	audio_pes->next = NULL;
	return(m_audio_pes);

}




/* Test function */
audio_pes_t * print_it(audio_pes_t * audio_pes){

	audio_pes_t * save;

	save = audio_pes->next;
	save->previous = NULL;

	printf("Time: %llu \n", audio_pes->time);
	free(audio_pes);
	return(save);
}



audio_pes_t * get_last_audio_pes(){


	audio_pes_t * tmp_audio_pes;
	audio_pes_t * audio_pes;

	/* decrease the audio pes counter */
	nr_audio_pes--;
	
	if(m_audio_pes == NULL) {
		if(debug >= 1) {
			printf("Get_Last_Audio_Pes: The main m_audio_pes == NULL exiting\n");
		}
		exit(-1);
	}

	tmp_audio_pes = m_audio_pes;

	while (tmp_audio_pes->next) {
		tmp_audio_pes = tmp_audio_pes->next;
	}

	tmp_audio_pes = tmp_audio_pes->previous;
	audio_pes = tmp_audio_pes->next;
	tmp_audio_pes->next = NULL;
	audio_pes->previous = NULL;
	return(audio_pes);
}


/* Test function to find a bug in the mpeg-II audio */

void find_error(audio_pes_t * audio_pes){

	uint32_t k;
	uint8_t * pt1;
	uint8_t comp[]={0x00, 0x00, 0x01, 0xc0};
	int gotit=0;

	pt1 = audio_pes->pes_data;


	for(k=0; k < audio_pes->size - (4 + 1); k++, pt1++) {
		if (memcmp(pt1, comp, 4) == 0) {
			printf("Found it at offset %i, chunk is %i, time is %llu \n", pt1 - audio_pes->pes_data, chunk_nr - 2, audio_pes->time);
			gotit=1;
			break;
		}
	}

	if(gotit){
		pt1 = audio_pes->pes_data;
		printf("\n");
		for( k=0; k <audio_pes->size && k < 40; k++, pt1++){
			printf(" %02x", pt1[0]);
			if((k+1)%16 == 0){
				printf("\n");
			}
		}
		printf("\n");
	}

}



void parse_audio(chunk_t * chunks) {



	/* Interation */
	int i, t, k;
	uint32_t p;

	/* Markers */
	int gotit;
	int gotnextaudio;
	int drop_pes;
	int foundit;

	/*Size in the chunk*/
	int size;
	int tsize;

	/* Offsets */
	uint32_t pes_header_offset;
	int offset_size;
	int data_offset;

	/* Audio overlap sizes */
	int record_size;
	int record_offset;


	/* Audio PES */
	audio_pes_t * audio_pes;
	uint8_t * pes_header;
	uint8_t * pes_data=NULL;
	uint8_t * tmp_pes_data=NULL;

	/* Pointers to magic with */
	uint8_t * pt1, * pt2;

	if(debug >= 3) {
		printf("Parse Audio: chunk %i\n", chunk_nr - 2);
	}

	i=0;
	if(audio_pes_not_finished){

		if(debug >= 6) {
			printf("Parse Audio: overlap from chunk %i \n", chunk_nr - 1);
		}

		audio_pes = get_last_audio_pes();

		/* Init variables */
		gotit = 0;
		size = audio_pes->size;
		pes_data=audio_pes->pes_data;
		record_offset = audio_pes->pes_header_record_offset;
		record_size = audio_pes->pes_header_record_size;
		data_offset = 0;
		drop_pes = 0;

		/* Okay we had a pes overlap, the header is no worry
		it's allready taken care of but how about the data :(
		Well the last pes record header could be small and overlap into
		this record. Hence we will check out if our stored record size is zero
		if it is then we know that the pes header may lap into our
		data record. We now check the pes_header record size and the offset
		to the start of the pes header. This will tell us how much
		over lap we have. */

		if(size == 0) {
			/* The actual pes header record size is the size of
			that record minus the offset to the start of the pes
			header.This will form our data_offset header_size is the size
			of the pes header */
			data_offset = header_size - (record_size - record_offset);
		}

		for(; i < chunks->nr_records; i++) {
			/* Okay copy all Audio data of type
			0x2c0 and 0x4c0 into the pes. Since I only have a
			DTivo I dunno about what the difference is between the
			to audio records.

			Since we can have over run of pes data to the next
			pes header (0x9c0 or 0x3c0) we will need to treat
			that one differntly that we did with the video pes.

			The rest is either video or junk (CC,TT EE type records)

			We are doing this until we find the next audio pes or
			the end of the chunk. If we find the end of the chunk
			we flag the audio_pes_not_finished unless it's the last
			chunk in our file. If so we drop the pes */

			switch(chunks->record_header[i].type){
				case 0x2c0:
				case 0x4c0:
					tsize = size;
					size = size + chunks->record_header[i].size - data_offset;
					audio_pes->size = size;
					pes_data = (uint8_t *)realloc(pes_data, size * sizeof(uint8_t));
					pt1 = pes_data + tsize;
					pt2 = chunks->records[i].record + data_offset;
					memcpy(pt1, pt2,
						sizeof(uint8_t) * chunks->record_header[i].size);
					audio_pes->pes_data = pes_data;
					break;
				case 0x3c0:
				case 0x9c0:
					gotit=1;
					break;
			}
			if(gotit == 1){
				audio_pes_not_finished = 0;
				break;
			}

		}

		if(gotit){
			size = chunks->record_header[i].size;
			tmp_pes_data = (uint8_t *)malloc(sizeof(uint8_t) * chunks->record_header[i].size);
			memcpy(tmp_pes_data,chunks->records[i].record,sizeof(uint8_t) * chunks->record_header[i].size);

			/* Sinse audio pes headers can be small and the fact that
			they not always start at 0 offset copy the next audio record
			if our record it's smaller than 40 (40 is just a good guess)*/

			if(size < 40 ) {
				gotnextaudio=0;
				/* Now get the next audio record */
				for(k=i+1; k < chunks->nr_records; k++) {
					switch(chunks->record_header[k].type){
						case 0x2c0:
						case 0x3c0:
						case 0x4c0:
						case 0x9c0:
							tmp_pes_data = (uint8_t *)realloc(tmp_pes_data,sizeof(uint8_t)
								* (chunks->record_header[i].size + chunks->record_header[k].size));
							pt1 = tmp_pes_data + size;
							memcpy(pt1,chunks->records[k].record,
								sizeof(uint8_t) * chunks->record_header[k].size);
							gotnextaudio = 1;
							size = size + chunks->record_header[k].size;
							break;
					}
					if(gotnextaudio) {
						break;
					}
				}

				if(!gotnextaudio) {
					/* bummer we didn't find the next audio
					record in this chunk :(. Hence we need to
					search for it in the next chunk if we have one.*/

					if(chunks->next != NULL) {
						for( k=0; k < chunks->next->nr_records; k++) {
							switch(chunks->next->record_header[k].type){
								case 0x2c0:
								case 0x3c0:
								case 0x4c0:
								case 0x9c0:
									tmp_pes_data = (uint8_t *)realloc(tmp_pes_data,sizeof(uint8_t)
										* (chunks->record_header[i].size + chunks->next->record_header[k].size));
									pt1 = tmp_pes_data + size;
									memcpy(pt1,chunks->next->records[k].record,
										sizeof(uint8_t) * chunks->next->record_header[k].size);
									gotnextaudio = 1;
									size = size + chunks->next->record_header[k].size;
									break;
							}
							if(gotnextaudio) {
								break;
							}
						}
					} else {
						if(debug >= 1) {
							printf("Parse Audio: pes drop - didn't find next audio pes overlap  - chunk %i\n", chunk_nr - 2);
						}
						drop_pes=1;
					}
				}
			}

			/* Now search for the start of the pes header */
			foundit = 0;
			pt1 = tmp_pes_data;
			/* FIXME need to check if we should habe header_size or (header_size + 1) since the size of the header is header_size */
			for(k=0; k < size - (header_size + 1); k++, pt1++) {
				if (memcmp(pt1, pes_a, 4) == 0) {
					foundit=1;
					break;
				}
			}



			/* Reset the size of the data record */
			size = 0;

			if(foundit) {

				/* Now we will need to see how much we steped fwd in the
				record. NOTE since we never will use data from the next
				record this is quite simple.

				The thing is that we really don't care where the next pes
				header is located since we will search for it's start anyway.

				This eliminates the trouble of storing offsets and similar
				thing between chunks
				*/


				offset_size = pt1 - tmp_pes_data;
				size = audio_pes->size;
				audio_pes->pes_data = (uint8_t *)realloc(audio_pes->pes_data,
					sizeof(uint8_t) * (size + offset_size));
				pt1 = audio_pes->pes_data;
				pt1 = pt1 + size;
				memcpy(pt1, tmp_pes_data, sizeof(uint8_t) * offset_size);
				audio_pes->size = audio_pes->size + offset_size;

			} else {
				/* Bummer we didn't get the pes end :( */
				if(debug >= 1) {
					printf("Parse Audio: pes end not found - audio pes overlap 2 - chunk %i\n", chunk_nr - 2);
				}
				audio_pes_not_finished=0;
				drop_pes=1;
			}

			/* Free our tmp_pes_data buffer */
			free(tmp_pes_data);
		}


		if(debug >= 1 && m_audio_pes != NULL && gotit == 1) {
			if(audio_pes->size != audio_pes_size && audio_pes_not_finished == 0) {
				printf("Parse Audio: pes size error - %i  - audio pes overlap - chunk %i \n", audio_pes->size, chunk_nr);
				printf("Parse Audio: Pes Header: ");
				for(t=0; t < header_size ; t++){
					printf(" %02x", audio_pes->pes_header[t]);
				}
				printf("\n");
			}
		}

		if(m_audio_pes != NULL){
			if(drop_pes != 1) {
				m_audio_pes = insert_audio_pes(audio_pes);
			}
		} else {
			if(drop_pes != 1) {
				m_audio_pes = audio_pes;
			}
		}

		/* It's like this if we don't find the next pes header in this chunk i.e. gotit == 0
		(not foundit) then we still haven't finished the audio pes and we set it to not finsihed
		and start over again. */
		if(gotit == 0) {
			/* This mean that we reach the end of the chunk but
			didn't find the end of our present audio pes. We now rise
			the audio_pes_not_finished flag unless it's the last chunk */
			if(chunks->next != NULL) {
				audio_pes_not_finished = 1;
			} else {
				audio_pes_not_finished = 0;
			}
		}
	}


/*******************************************************
********************************************************/



	for(;i < chunks->nr_records; i++) {

		if(debug >= 6) {
			printf("Parse Audio: record %i of %i records - chunk %i\n",i + 1, chunks->nr_records, chunk_nr -2 );
		}

		switch(chunks->record_header[i].type){
			case 0x9c0:
			case 0x3c0:
				if(debug >= 6){
					printf("Parse Audio: got pes header - record %i \n", i + 1);
				}

				/* Allocate and init the pes structure */
				audio_pes = (audio_pes_t *)malloc(sizeof(audio_pes_t));
				audio_pes->pes_data=NULL;
				audio_pes->pes_header=NULL;
				audio_pes->next=NULL;
				audio_pes->previous=NULL;
				audio_pes->time = 0;
				audio_pes->size = 0;
				/* Used if we have a pes overlap to the next chunk */
				audio_pes->pes_header_record_size = chunks->record_header[i].size;
				audio_pes->pes_header_record_offset = 0;

				/* Tmp holder of pes header data */
				tmp_pes_data=NULL;

				/* Reset the pes_data */
				pes_data=NULL;

				/* Rest the offset to the next record */
				pes_header_offset = 0;

				/* Maker */
				drop_pes = 0;
				gotnextaudio=0;

				/* Now the fact is that Audio pes header in tivo
				can be small so it doesn't hold the whole pes header
				but just a fraction of it and the rest continues
				in the next Audio data record :(

				Hence we need to check copy the next audio recor
				to the end of this record if the size of this
				record is 40 (calculated guess) or less and
				then do a search for the pes start since
				the pes header isn't always starting on the
				firt byte in the record :(. */


				size = chunks->record_header[i].size;
				tmp_pes_data = (uint8_t *)malloc(sizeof(uint8_t) * chunks->record_header[i].size);
				memcpy(tmp_pes_data,chunks->records[i].record,sizeof(uint8_t) * chunks->record_header[i].size);



				if(size < 40 ) {
					if(debug >= 7) {
						printf("Parse Audio: small pes audio record size %i < 40\n",size);
					}
					/* Now get the next audio record */
					for(k=i+1; k < chunks->nr_records; k++) {
						switch(chunks->record_header[k].type){
							case 0x2c0:
							case 0x3c0:
							case 0x4c0:
							case 0x9c0:
								tmp_pes_data = (uint8_t *)realloc(tmp_pes_data,sizeof(uint8_t)
									* (chunks->record_header[i].size + chunks->record_header[k].size));
								if(debug >= 7) {
									printf("Parse Audio: pes audio size %i next %i\n",chunks->record_header[i].size,chunks->record_header[k].size);
								}
								pt1 = tmp_pes_data + size;
								memcpy(pt1,chunks->records[k].record,
									sizeof(uint8_t) * chunks->record_header[k].size);
								gotnextaudio = 1;
								size = size + chunks->record_header[k].size;
								break;
						}
						if(gotnextaudio) {
							break;
						}
					}

					if(!gotnextaudio) {
						/* bummer we didn't find the next audio
						record in this chunk :(. Hence we need to
						search for it in the next chunk if we have one.*/
						if(debug >= 7) {
							printf("Parse Audio: next pes in next chunk\n");
						}
						if(chunks->next != NULL) {
							for( k=0; k < chunks->next->nr_records; k++) {
								switch(chunks->next->record_header[k].type){
									case 0x2c0:
									case 0x3c0:
									case 0x4c0:
									case 0x9c0:
										tmp_pes_data = (uint8_t *)realloc(tmp_pes_data,sizeof(uint8_t)
											* (chunks->record_header[i].size + chunks->next->record_header[k].size));
										pt1 = tmp_pes_data + size;
										memcpy(pt1,chunks->next->records[k].record,
											sizeof(uint8_t) * chunks->next->record_header[k].size);
										gotnextaudio = 1;
										size = size + chunks->next->record_header[k].size;
										break;
								}
								if(gotnextaudio) {
									break;
								}
							}
						} else {
							/* Drop this pes it was the last one
							in the file we converted FIXME */
							drop_pes=1;
						}
					}
				}

				/* Now search for the start of the pes header */
				gotit = 0;
				pt1 = tmp_pes_data;
				/* FIXME need to check if we should habe header_size or (header_size + 1) since the size of the header is header_size */
				for(k=0; k < size - (header_size + 1); k++, pt1++) {
					if (memcmp(pt1, pes_a, 4) == 0) {
						gotit=1;
						audio_pes->pes_header_record_offset = pt1 - tmp_pes_data;
						break;
					}
				}


				/* Reset the size of the data record */
				size = 0;

				if(gotit) {
					/* Pes header of AC3 is header_size bytes FIXME */
					pes_header = (uint8_t *)malloc(sizeof(uint8_t) * header_size);
					memcpy(pes_header, pt1, sizeof(uint8_t) * header_size);
					audio_pes->pes_header = pes_header;

					/* Now we will need to see how much we steped fwd in the
					record and if we used any bytes from the next record. */
					if(gotnextaudio) {
						pes_header_offset = (pt1 + header_size) - tmp_pes_data;
						if(pes_header_offset >= chunks->record_header[i].size) {
							/* So we used data from the next record
							but how much did we use?*/
							pes_header_offset = pes_header_offset - chunks->record_header[i].size;
						} else {
							/* There might be data beloning to the header in this record
							happens with slightly bigger size of the record i.e. header_size and above */
							pt1 = pt1 + header_size;
							pes_data = (uint8_t *)malloc(sizeof(uint8_t) *
									(chunks->record_header[i].size - pes_header_offset));
							memcpy(pes_data, pt1, chunks->record_header[i].size - pes_header_offset);
							size = chunks->record_header[i].size - pes_header_offset;
							audio_pes->size = size;
							audio_pes->pes_data = pes_data;
							pes_header_offset = 0;
						}
					} else {

						/* Location of the start of the pes package in
						the record. The if is to see if we had any data
						present after the header and if so copy it*/
						pes_header_offset = (pt1 + header_size) - tmp_pes_data;
						if(pes_header_offset <= chunks->record_header[i].size) {
							pt1 = pt1 + header_size;
							pes_data = (uint8_t *)malloc(sizeof(uint8_t) *
								(chunks->record_header[i].size - pes_header_offset));
							memcpy(pes_data, pt1, chunks->record_header[i].size - pes_header_offset);
							size = chunks->record_header[i].size - pes_header_offset;
							audio_pes->size = size;
							audio_pes->pes_data = pes_data;
							if(debug >= 7) {
								printf("Parse Audio: pes audio header record contains data\n Parse Audio: Data:");
								for(p=0; p < audio_pes->size && p < 16; t++){
									printf(" %02x", audio_pes->pes_data[p]);
								}
								printf("\n");
							}

						}
						pes_header_offset = 0;
					}

					if(debug >= 6) {
						printf("Parse Audio: header: ");
						for(k=0; k < header_size; k++){
							printf(" %02x", pes_header[k]);
						}
						printf("\n");
					}

					/* Find PTS time of PES */
					pt1 = audio_pes->pes_header;
					pt1 = pt1 + time_offset;
					audio_pes->time = get_time(pt1);
					if(audio_start == 0) {
						audio_start = audio_pes->time;
						if(debug >= 3) {
							printf("Parse Audio: Audio start at %llu ticks\n", audio_start);
						}
					}
					if(debug >= 6){
						printf("Parse Audio: PTS (ticks): %llu\n",audio_pes->time);
					}
				} else {
					if(debug >= 1){
						printf("Parse Audio: dropped pes didn't find header start - %i of %i records - chunk %i\n",i + 1, chunks->nr_records, chunk_nr);
					}
					drop_pes=1;
					free(audio_pes);
				}

				/* Free our tmp_pes_data buffer */
				free(tmp_pes_data);


				/*****************************************/


				/* Get the pes payload, this is somewhat started
				above but we must find the next pes header to be sure
				we got all data*/


				if(!drop_pes) {
					if(debug >= 6){
						printf("Parse Audio: pes audio payload - %i of %i records - chunk %i\n",i + 1, chunks->nr_records, chunk_nr);
					}

					/* Find end of the audio PES */

					gotit=0;
					for(i=i+1; i < chunks->nr_records; i++) {
						/* Okay copy all Audio data of type
						0x2c0 and 0x4c0 into the pes (I only have
						a DTivo so when I get SA samples this needs to be checked).

						Since we can have over run of pes data to the next
						pes header record (0x9c0 or 0x3c0) we will need to treat
						that one differntly that we did above the video.

						The rest is either video or junk (CC,TT EE type records)

						We are doing this until we find the next audio pes or
						the end of the chunk. If we find the end of the chunk
						we flag the audio_pes_not_finished unless it's the last
						chunk in our file. If so we drop the pes */

						switch(chunks->record_header[i].type){
							case 0x2c0:
							case 0x4c0:
								tsize = size;
								size = size + chunks->record_header[i].size - pes_header_offset;
								audio_pes->size = size;
								pes_data = (uint8_t *)realloc(pes_data, size * sizeof(uint8_t));
								pt1 = pes_data + tsize;

								pt2 = chunks->records[i].record + pes_header_offset;
								memcpy(pt1, pt2,
									(sizeof(uint8_t) * chunks->record_header[i].size) - pes_header_offset);
								audio_pes->pes_data = pes_data;
								if(debug >= 7) {
									printf("Parse Audio: data: \n");
									for(t=0; t < size && t < 16; t++) {
										printf(" %02x", pes_data[t]);
									}
									printf("\n");
								}
								break;
							case 0x3c0:
							case 0x9c0:
								gotit=1;
								break;
						}
						if(gotit == 1){
							break;
						}

					}

					/* So we got the next pes header - now the fact is that we can have
					overuns from the previous pes into this pes header record :(. Hence
					what we need to do is to find the start of the pes header and copy all
					the data before that into the pes we are dealing with we will in other
					words repeate the same thing from above header but in reverse so to say :(*/

					if(gotit){
						size = chunks->record_header[i].size;
						tmp_pes_data = (uint8_t *)malloc(sizeof(uint8_t) * chunks->record_header[i].size);
						memcpy(tmp_pes_data,chunks->records[i].record,sizeof(uint8_t) * chunks->record_header[i].size);

						/* Sinse audio pes headers can be small and the fact that
						they not always start at 0 copy the next audio record
						if our record it's smaller than 40 */

						if(size < 40 ) {
							gotnextaudio=0;
							/* Now get the next audio record */
							for(k=i+1; k < chunks->nr_records; k++) {
								switch(chunks->record_header[k].type){
									case 0x2c0:
									case 0x3c0:
									case 0x4c0:
									case 0x9c0:
										tmp_pes_data = (uint8_t *)realloc(tmp_pes_data,sizeof(uint8_t)
											* (chunks->record_header[i].size + chunks->record_header[k].size));
										pt1 = tmp_pes_data + size;
										memcpy(pt1,chunks->records[k].record,
											sizeof(uint8_t) * chunks->record_header[k].size);
										gotnextaudio = 1;
										size = size + chunks->record_header[k].size;
										break;
								}
								if(gotnextaudio) {
									break;
								}
							}

							if(!gotnextaudio) {
								/* bummer we didn't find the next audio
								record in this chunk :(. Hence we need to
								search for it in the next chunk if we have one.*/

								if(chunks->next != NULL) {
									for( k=0; k < chunks->next->nr_records; k++) {
										switch(chunks->next->record_header[k].type){
											case 0x2c0:
											case 0x3c0:
											case 0x4c0:
											case 0x9c0:
												tmp_pes_data = (uint8_t *)realloc(tmp_pes_data,sizeof(uint8_t)
													* (chunks->record_header[i].size + chunks->next->record_header[k].size));
												pt1 = tmp_pes_data + size;
												memcpy(pt1,chunks->next->records[k].record,
													sizeof(uint8_t) * chunks->next->record_header[k].size);
												gotnextaudio = 1;
												size = size + chunks->next->record_header[k].size;
												break;
										}
										if(gotnextaudio) {
											break;
										}
									}
								} else {
									/* Drop this pes it was the last one
									in the file we converted */
									drop_pes = 1;
									if(debug >= 1) {
										printf("Parse Audio: pes audio dropped - %i of %i records - chunk %i\n",i + 1, chunks->nr_records, chunk_nr);
									}
								}

							}
						}

						/* Now search for the start of the pes header */
						foundit = 0;
						pt1 = tmp_pes_data;
						/* FIXME need to check if we should habe header_size or (header_size + 1) since the size of the header is header_size */
						for(k=0; k < size - (header_size + 1); k++, pt1++) {
							if (memcmp(pt1, pes_a, 4) == 0) {
								foundit=1;
								break;
							}
						}



						/* Reset the size of the data record */
						size = 0;

						if(foundit) {

							/* Now we will need to see how much we steped fwd in the
							record. NOTE since we never will use data from the next
							record this is quite simple.

							The thing is that we really don't care where the next pes
							header is located since we will search for it's start anyway.

							This eliminates the trouble of storing offsets and similar
							thing between chunks
							*/


							offset_size = pt1 - tmp_pes_data;

							if(debug >= 6) {
								printf("Parse Audio: pes audio data - next pes header - %i bytes is overlapping\n", offset_size);
							}

							size = audio_pes->size;
							audio_pes->pes_data = (uint8_t *)realloc(audio_pes->pes_data,
								sizeof(uint8_t) * (size + offset_size));
							pt1 = audio_pes->pes_data;
							pt1 = pt1 + size;
							memcpy(pt1, tmp_pes_data, sizeof(uint8_t) * offset_size);
							audio_pes->size = audio_pes->size + offset_size;


						} else {
							/* Bummer we didn't get it FIXME */
							if(debug >= 1) {
								printf("Parse Audio: pes audio dropped didn't find end of pes - %i of %i records - chunk %i\n",i + 1, chunks->nr_records, chunk_nr);
							}
							drop_pes=1;
							audio_pes_not_finished=0;
							if(audio_pes->pes_data != NULL){
								free(audio_pes->pes_data);
							}
							free(audio_pes->pes_header);
							free(audio_pes);
						}

						/* Free our tmp_pes_data buffer */
						free(tmp_pes_data);

						/* reset i so the catch the next pes */
						i--;
					} else {
						audio_pes_not_finished = 1;
					}

				}

				if(!drop_pes){
					if(m_audio_pes != NULL && debug >= 1){
						if(audio_pes->size != audio_pes_size && audio_pes_not_finished == 0) {
							printf("Parse Audio: pes size error %i - audio pes - chunk %i \n", audio_pes->size, chunk_nr);
							printf("Parse Audio: Header: ");
							for(t=0; t < header_size ; t++){
								printf(" %02x", audio_pes->pes_header[t]);
							}
							printf("\n");
						}
					}

					if(m_audio_pes != NULL){
						//find_error(audio_pes);
						m_audio_pes = insert_audio_pes(audio_pes);

					} else {
						m_audio_pes = audio_pes;
					}
				}

				/* Hmm not a clear function  - well if gotit == 0 it means that we didn't find the next pes header
				hence we will need to seach for it in the next chunk (note the difference between gotit and foundit)*/

				if(gotit == 0 && drop_pes != 1) {
					/* This mean that we reach the end of the chunk but
					didn't find the end of our present audio pes. We now rise
					the audio_pes_not_finished flag unless it's the last chunk */
					if(chunks->next != NULL) {
						audio_pes_not_finished = 1;
					} else {
						/* Well there will be no next chunk so it really doesn't matter */
						audio_pes_not_finished = 0;
					}
				}
				break;
		}
	}

}
