/*
 * Copyright (C) 2002  Olaf Beck <olaf_sc@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "tydemux.h"
#include "global.h"


void bsort(uint64_t video_time_array[], int size) {

	uint64_t tmp;
	int i, j;

	for(i=0; i < size; i++) {
		for(j=0; j < size; j++) {
			if(video_time_array[i] < video_time_array[j]) {
				tmp = video_time_array[i];
				video_time_array[i] = video_time_array[j];
				video_time_array[j] = tmp;
			}
		}
	}

}


void print_chunk(chunk_t * chunk) {

	/* Iteration */
	int i;
	uint32_t j;

	for (i=0; i < chunk->nr_records; i++) {
		printf("Record: %i size: %u type: %03x\n", i + 1, chunk->record_header[i].size, chunk->record_header[i].type);
		if(chunk->record_header[i].size != 0) {
			printf("Record data: ");
			for (j=0; j < chunk->record_header[i].size && j < 16; j ++) {
				printf("%02x ", chunk->records[i].record[j]);
			}
			printf("\n");
		}
	}

}

void free_chunk(chunk_t * chunk) {

	/* Iteration */
	int i;

	for (i=0; i < chunk->nr_records; i++) {
		if(chunk->record_header[i].size != 0) {
			free(chunk->records[i].record);
		}
	}
	free(chunk->records);
	free(chunk->record_header);
	free(chunk);
}



chunk_t * add_chunk(chunk_t * chunk, chunk_t * chunks){

	/* Iteration */
	int i;

	/* The first chunk */
	chunk_t * save;

	/* If we want to save new
	duplicated chunk
	chunk_t * old_chunk;*/

	/* Uplication indicator */
	int duplicate;

	extern int realchunk_nr;

	save = chunks;
	duplicate = 0;

	while (chunks->next) {
		chunks = chunks->next;
	}

	/* Check for possible duplication of chunk */
	if(chunk->nr_records == chunks->nr_records && chunk->unknown == chunks->unknown) {
		/* We might have a duplicate :( */
		for(i=0; i < chunks->nr_records; i++) {
			if(chunks->record_header[i].size == chunk->record_header[i].size &&
			   chunks->record_header[i].type == chunk->record_header[i].type &&
			   chunks->record_header[i].junk1 == chunk->record_header[i].junk1 &&
			   chunks->record_header[i].junk2 == chunk->record_header[i].junk2 &&
			   chunks->record_header[i].junk3 == chunk->record_header[i].junk3) {
				duplicate = 1;
			} else {
				duplicate = 0;
				break;
			}
		}
	}

	if (duplicate) {
		/* Well either we keep the old chunk
		 or we replace it with the new one -
		 good guees is that the new one is
		 better than the old one "error"
		 made them rerecord it??*
		 Anyways still sticking with the old chunk.
		 */
		if (debug >= 2) {
			printf("Dupliacted chunk %i - skipping\n", chunk_nr);
		}
		free_chunk(chunk);
		realchunk_nr--;
		/*Keep new chunk code
		if (debug >= 2) {
			printf("Dupliacted chunk %i - skipping\n", chunk_nr);
		}
		old_chunk = chunks;
		chunk->previous = chunks->previous;
		chunk->next = null;
		chunk->previous->next = chunk;
		free_chunk(old_chunk);
		*/
	} else {
		chunks->next = chunk;
		chunk->previous = chunks;
		chunk->next = NULL;
	}
	return(save);
}



/* FIXME This one should not be used!! */

void free_chunks(chunk_t * chunks) {

	int i;
	chunk_t * old_chunk;

	while(chunks->next) {
		old_chunk = chunks;
		chunks = chunks->next;
		for(i=0; i < old_chunk->nr_records; i++) {
			free(old_chunk->records[i].record);
		}
		free(old_chunk->record_header);
		free(old_chunk->records);
		free(old_chunk);
	}

	for(i=0; i < chunks->nr_records; i++) {
		free(chunks->records[i].record);
	}
	free(chunks->record_header);
	free(chunks->records);
	free(chunks);
}



chunk_t * parse_chunk(chunk_t * chunks) {


	/* Chunks */
	chunk_t * old_chunk;

	if (debug >= 3) {
		printf("Parse Chunk: %i\n", chunk_nr - 2);
	}

	parse_video(chunks);
	parse_audio(chunks);


	old_chunk = chunks;
	chunks = chunks->next;
	chunks->previous = NULL;
	free_chunk(old_chunk);
	realchunk_nr--;
	return(chunks);

}

/*
record_header_t * oread_record(int nr_records, int file_nr) {

	int i;

	record_header_t * record_header;
	tmp_record_header_t * tmp_record_header;

	record_header = (record_header_t *)malloc(sizeof(record_header_t) * nr_records);
	tmp_record_header = (tmp_record_header_t *)malloc(sizeof(tmp_record_header_t) * nr_records);

	read(file_nr, tmp_record_header,sizeof(tmp_record_header_t) * nr_records);

	if(debug >= 4) {
		printf("Record Header Read: chunk %i\n", chunk_nr);
	}

	for(i=0; i < nr_records; i++) {
		if((tmp_record_header[i].size1 >> 4) == 0x08) {
			record_header[i].size = 0;
		} else {
			record_header[i].size =  (tmp_record_header[i].size1 << 12)
						+(tmp_record_header[i].size2 << 4)
						+(tmp_record_header[i].size3);
		}
		*//* I don't get it why I coudn't get bswap64 to work ?? *//*
		record_header[i].type =  (tmp_record_header[i].type1 << 8)
					+(tmp_record_header[i].type2);
		record_header[i].junk1 = B2N_32(tmp_record_header[i].junk1);
		record_header[i].junk2 = B2N_32(tmp_record_header[i].junk2);
		record_header[i].junk2 = B2N_32(tmp_record_header[i].junk3);
		if(debug >= 4) {
			printf("\nRecord: %i\n", i+1);
			printf("Record size %u\n", record_header[i].size);
			printf("Record type %03x\n", record_header[i].type);
			printf("Record junk1 %08x\n", record_header[i].junk1);
			printf("Record junk2 %08x\n", record_header[i].junk2);
			printf("Record junk3 %08x\n\n", record_header[i].junk3);
		}
	}
	free(tmp_record_header);
	return(record_header);
}
*/

record_header_t * read_record(int nr_records, int file_nr) {

	int i, j;

	record_header_t * record_header=NULL;
	uint8_t * tmp_record_array=NULL;

	record_header = (record_header_t *)malloc(sizeof(record_header_t) * nr_records);
	tmp_record_array = (uint8_t *)malloc(sizeof(uint8_t) * nr_records * 16);
	read(file_nr, tmp_record_array,sizeof(uint8_t) * nr_records * 16);

	if(debug >= 4) {
		printf("Record Header Read: chunk %i, records %i\n", chunk_nr, nr_records);
	}

	if(debug >= 8) {
		for(i=0; i < nr_records * 16; i = i + 16) {
			for(j=0; j < 16 ; j++ ) {
				printf(" %02x", tmp_record_array[i + j]);
			}
			printf("\n");
		}
	}


	for(i=0, j=0; i < nr_records * 16; j++, i = i + 16) {

			if((tmp_record_array[i] >> 4) == 0x08) {
					record_header[j].size = 0;
			} else {
				/* Get record size */
				record_header[j].size = (tmp_record_array[i] << 12)
										+(tmp_record_array[i+1] << 4)
										+(tmp_record_array[i+2] >> 4);
			}

			/* Get record type */
			record_header[j].type = ((tmp_record_array[i+2] & 0x0f) << 8)
									+(tmp_record_array[i+3]);
			/* The rest of the junk - need to figure out what it's all
			about since it's not a real pes header or anything close */
			record_header[j].junk1 = (tmp_record_array[i+4] << 24)
									+(tmp_record_array[i+5] << 16)
									+(tmp_record_array[i+6] << 8)
									+(tmp_record_array[i+7]);

			record_header[j].junk2 = (tmp_record_array[i+8] << 24)
									+(tmp_record_array[i+9] << 16)
									+(tmp_record_array[i+10] << 8)
									+(tmp_record_array[i+11]);

			record_header[j].junk3 = (tmp_record_array[i+12] << 24)
									+(tmp_record_array[i+13] << 16)
									+(tmp_record_array[i+14] << 8)
									+(tmp_record_array[i+15]);

			if(debug >= 4) {
				printf("\nRecord: %i\n", j+1);
				printf("Record size %u\n", record_header[j].size);
				printf("Record type %03x\n", record_header[j].type);
				printf("Record junk1 %08x\n", record_header[j].junk1);
				printf("Record junk2 %08x\n", record_header[j].junk2);
				printf("Record junk3 %08x\n\n", record_header[j].junk3);
			}
	}
	free(tmp_record_array);
	return(record_header);
}


chunk_t * check_chunk(chunk_t * chunk) {

	/* iteration */
	int i, j;

	/* Video time array */
	uint64_t * tmp_video_time_array;

	/* Video time ticks */
	uint64_t first_tick;
	uint64_t tmp_max_ticks;
	uint64_t tmp_min_ticks;

	/* Video sync okay or not */
	int okay;


	/* Nr of video element */
	int nr_video;

	/* Pointers to do magic with */
	uint8_t * pt1;



	/**
	 * Audio check
	 *
	 * If we have a TyStream that has e.g. AC3 audio then
	 * a chunk with MPEG-II audio is definitly a junk chunk.
	 * This can happen in the start of a recording but
	 * also in the middle of a recording when a OOS is
	 * encountered. OOS == Out Of Sync i.e. a chunk
	 * which is not in sync with the rest of the TyStream
	 * usually from an other PID in the DirectTV PID - or
	 * just junk in general
	 *
	 *
	 * FIXME May have bad impact on shows how alter ac3/mpeg-II
	 * We may want to set some form of indicator for this.
	 */
	if (debug >= 3){
		printf("Check Chunk: %i\n", chunk_nr);
	}

	for(i=0; i < chunk->nr_records; i++){
		if(chunk->record_header[i].type == wrong_audio) {
			if(debug >= 2) {
				printf("Check Chunk: %i Record %i - wrong type of audio - skipping\n", chunk_nr, i + 1);
			}
			return(0);
		}
	}


	/**
	 * Junk record type check
	 *
	 * I have found that we sometimes have junk records types
	 * in the chunk if we have a garbage chunk.
	 */

	for(i=0; i < chunk->nr_records; i++){
		switch(chunk->record_header[i].type) {
			case 0x2c0: /* Audio Data */
			case 0x3c0: /* Audio PES - MPEG */
			case 0x4c0: /* Audio Data */
			case 0x9c0: /* Audio PES - AC3 */
			case 0x2e0: /* Video Data */
			case 0x6e0: /* Video PES */
			case 0x7e0: /* Video SEQ */
			case 0x8e0: /* Video I Frame/Field */
			case 0xae0: /* Video P Frame/Field */
			case 0xbe0: /* Video B Frame/Field */
			case 0xce0: /* Video GOP */
			case 0xe01: /* Video CC  (ClosedCaptions) */
			case 0xe02: /* Video EE  (Extended Data)*/
			case 0xe05: /* Video TT (TeleText) */
				break;
			default:
				if(debug >= 2) {
					printf("Check Chunk: chunk %i record: %i - junk record type: %03x  - skipping\n",
						chunk_nr, i + 1, chunk->record_header[i].type);
				}
				return(0);
				break;

		}
	}

	/**
	 * Video / Audio in sync check
	 *
	 * So we passed the checks above - lets now see if we
	 * have Video and Audio in sync i.e. that our PTS
	 * for Video doesn't differ more than 4xmax_ticks from our
	 * previous chunk last PTS for Video - same check for audio
	 *
	 * This could be done in the parse chunk function but
	 * we will do it here - more over head but simpler in
	 * someways.
	 *
	 * Video first - Audio not yet implemented FIXME
	 *
	 * FIXME We don't do any fancy frame/field count which is
	 * nessesary if we want this to be safe.
	 */



	/* Fist how many Video PES headers */

	nr_video = 0;
	okay = 0;

	for(i=0; i<chunk->nr_records; i++) {
		switch(chunk->record_header[i].type) {
			case 0x6e0:
				nr_video++;
				break;
		}
	}

	if(nr_video != 0) {

		tmp_video_time_array = (uint64_t *)malloc(sizeof(uint64_t) * nr_video);

		for(i=0, j=0; i < chunk->nr_records; i++) {
			switch(chunk->record_header[i].type) {
				case 0x6e0:
					pt1 = chunk->records[i].record;
					pt1 = pt1 + pes_time_offset;
					tmp_video_time_array[j] = get_time(pt1);
					j++;
					break;
			}
		}



		if(nr_video > 1) {
			bsort(tmp_video_time_array, nr_video);
		}

		tmp_max_ticks = 0;
		tmp_min_ticks = 10000;

		if(nr_video > 2) {
			for(i=0; i < nr_video - 1; i++){
				if(tmp_max_ticks < (tmp_video_time_array[i+1] - tmp_video_time_array[i])) {
					tmp_max_ticks = tmp_video_time_array[i+1] - tmp_video_time_array[i];
				}
				if(debug >= 7) {
					printf("Check Chunk: Tick diff: %llu\n", tmp_video_time_array[i+1] - tmp_video_time_array[i]);
				}
				if(tmp_min_ticks > (tmp_video_time_array[i+1] - tmp_video_time_array[i])) {
					tmp_min_ticks = tmp_video_time_array[i+1] - tmp_video_time_array[i];
				}
			}
		}

		if(last_tick != 0) {
			first_tick = tmp_video_time_array[0];

			if(first_tick > last_tick) {

				if(first_tick - last_tick <= 4 * med_ticks) {
					/* We are okay - guess at the moment */
					okay = 1;
				} else if (first_tick - last_tick <= 10 * med_ticks) {
					/* FIXME need to align the sync of video/audio*/
					if(debug >= 1) {
						printf("Check Chunk: Warning!! Chunk: %i - hole in video stream\n", chunk_nr);
					}
					okay = 1;
				} else {
					if(debug >= 2) {
						printf("Check Chunk: %i - Out of Sync with %llu ticks - skipping\n", chunk_nr, first_tick - last_tick);
					}
				}

			} else if (first_tick < last_tick) {

				if(last_tick - first_tick <= 4 * med_ticks) {
					/* We are okay - guess at the moment */
					okay = 1;
				} else if (last_tick - first_tick <= 10 * med_ticks) {
					/* FIXME need to align the sync of video/audio*/
					if(debug >= 1) {
						printf("Check Chunk: Warning!! Chunk: %i - hole in video stream\n", chunk_nr);
					}
					okay = 1;
				} else {
					if(debug >= 2) {
						printf("Check Chunk: %i - Out of Sync with minus %llu ticks - skipping\n",chunk_nr, last_tick - first_tick);
					}
				}
			} else if (first_tick == last_tick) {
				if(debug >= 1) {
					printf("Check Chunk: Warning!! Chunk: %i - last/fist is equal - %llu\n", chunk_nr, last_tick);
				}
				/* We set okay here even if things are funky */
				okay = 1;
			} else {
				/* We set okay here even if things are funky */
				if(debug >= 1) {
					printf("Check Chunk: Warning!! Chunk: %i - this is not supposed to happen - %llu\n", chunk_nr, last_tick);
				}
				okay = 1;
			}

		} else {
			/**
			 * FIXME
			 * Okay first chunk to check we assume it's okay
			 * Will definitly need some check to see if all chunks
			 * after this one error out - i.e. the first chunk is bad
			*/
			okay = 1;
		}

		if(okay) {
			last_tick = tmp_video_time_array[nr_video -1];
			free(tmp_video_time_array);
			if(tmp_max_ticks != 0 && tmp_min_ticks != 10000 ) {
				med_ticks = (tmp_max_ticks + tmp_min_ticks)/2;
				if(debug >= 6) {
					printf("Check Chunk: med_ticks %llu - chunk %i\n", med_ticks, chunk_nr);
				}

			}
			return(chunk);
		} else {
			if(debug >= 2) {
				printf("Check Chunk: %i Out of sync - skipping\n", chunk_nr);
			}
			return(0);
		}

	} else {
		/* Bummer we need to check audio PTS sync :( FIXME */
		if(debug >= 1) {
			printf("Chunk Chunk: Warning!! chunk %i No video records - assume it's okay\n", chunk_nr);
		}
		return(chunk);
	}

}




chunk_t * read_chunk(int file_nr) {

	/* Interation */
	int i;

	/* The data part of the chunk */
	uint8_t * in_records;

	/* Chunk header 4 bytes */
	uint16_t records_nr;
	uint16_t unknown;


	/* Chunk related */
	record_header_t * record_header;
	chunk_t * chunk;
	record_t * records;

	read(file_nr, &records_nr, sizeof(uint16_t));
	read(file_nr, &unknown, sizeof(uint16_t));

	if(debug >= 3) {
		printf("Read Chunk: chunk %i nr_records: %i\n",chunk_nr, records_nr );
	}

	chunk = (chunk_t *)malloc(sizeof(chunk_t));
	chunk->nr_records = records_nr;
	chunk->unknown = unknown;

	records = (record_t *)malloc(sizeof(record_t) * chunk->nr_records);
	chunk->records = records;


	/* Check if we have a okay chunk or if it's just bogus */

	if (chunk->nr_records * 16 < chunk_size - 4) {

		record_header = read_record(records_nr, file_nr);
		chunk->record_header = record_header;
		for(i=0; i < chunk->nr_records; i++) {
			if (record_header[i].size == 0 ) {
				chunk->records[i].record = NULL;
			} else {
				in_records =(uint8_t *)malloc(sizeof(uint8_t) * record_header[i].size);
				read(file_nr, in_records,sizeof(uint8_t) * record_header[i].size);
				chunk->records[i].record = in_records;
			}
		}
		if(debug >= 5) {
			print_chunk(chunk);
		}
		return(chunk);

	} else {

		free(chunk->records);
		free(chunk);
		if(debug >= 2) {
			printf("Read Chunk: Bad chunk: %i - record overrun - skipping\n", chunk_nr);
		}
		return(0);

	}


}



