/*
 * Copyright (C) 2002  Olaf Beck <olaf_sc@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "sample.h"


int main(int argc, char *argv[]) {

	/* Args*/
	int flags;

	/* Switches */
	char * infile=NULL;
	char * outfile=NULL;

	/* Files */
	int file_nr;
	int out_file_nr;
	struct stat file_info;
	off_t file_size;

	/* Chunk things */
	int sample_chunks;
	int collected;
	int nr_of_samples;
	int nr_chunks;
	int probe_hz;

	int chunk_size;


	/* Copy buffer */
	uint8_t * buffer;

	/* Interation */
	int j;
	int chunk_nr;

	/* Init */
	chunk_size = 131072;

	/* Command line parsing */

	while ((flags = getopt(argc, argv, "?hi:o:")) != -1) {
		switch (flags) {
		case 'i':
			if(optarg[0]=='-') usage();
			infile = optarg;
			break;
		case 'o':
			if(optarg[0]=='-') usage();
			outfile = optarg;
			break;
		case '?':
			usage();
			break;
		case 'h':
			usage();
			break;
		default:
			usage();
		}
	}


	if (infile == NULL) {
		usage();
		exit(1);
	}

	if (outfile == NULL) {
		usage();
		exit(1);
	}

	 /* Open the TY file */
#ifdef WIN32
	file_nr = open(infile, O_RDONLY | _O_BINARY);
#else   /* Unix */
	file_nr = open(infile, O_RDONLY);
#endif

	if(file_nr == -1) {
		perror("");
		exit(-1);
	}

	fstat(file_nr, &file_info);
	file_size = file_info.st_size;



	/* Open output file */

#ifdef WIN32
	out_file_nr = open(outfile,O_WRONLY | O_CREAT | O_TRUNC | _O_BINARY, S_IREAD|S_IWRITE);
#else
	out_file_nr = open(outfile,O_WRONLY | O_CREAT | O_TRUNC, S_IREAD|S_IWRITE);
#endif


	if(out_file_nr == -1){
		perror("");
		exit(-1);
	}

	/* Get 10MB of data in each sample - i.e. sample_chunks == 1024 * 1024 * 10 / 131072 */
	sample_chunks = 80;

	/* We want 100MB of data */
	nr_of_samples = 10;

	if(file_size <= 104857600) {
		printf("Please send the whole file it's just %ldMB no idea to sample it", file_size / 1048576);
		exit (0);
	}

	nr_chunks = file_size / chunk_size;

	printf("Total number of chunks in file: %i\n", nr_chunks);

	/* How many chunks between each sample */
	probe_hz = nr_chunks / nr_of_samples;

	printf("Sample rate is %i chunks between samples\n", probe_hz);

	for(j=0, chunk_nr = 0, collected = 0; j < file_size && collected < nr_of_samples ; j = (j + chunk_size*probe_hz), chunk_nr = chunk_nr + probe_hz, collected++ ){

		printf("Sampling chunk %i to %i\n", chunk_nr, chunk_nr + 80);

		lseek(file_nr, chunk_nr * chunk_size, SEEK_SET);

		/* Grab 10MB of memory */
		buffer = (uint8_t *)malloc(sizeof(uint8_t) * 10485760);
		if(buffer == NULL) {
				printf("Memory allocation problem - exit\n");
				exit(-1);
		}

		if(read(file_nr, buffer, 10485760) != 10485760) {
				printf("Read problem - exit\n");
				exit(-1);
		}

		if(write(out_file_nr, buffer,10485760) != 10485760)  {
				printf("Write problem - exit\n");
				exit(-1);
		}
		free(buffer);
	}
	printf("Done\n");

	return(0);
}


void usage(){

	printf("\nsample 0.1 usage:\n"
	       "sample  -i infile -o outfile\n"
	       "\n");
	exit(1);
}
