/*
 * Copyright (C) 2002  Olaf Beck <olaf_sc@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "tydemux.h"



/* Global variables */
uint8_t * pes_a;

uint8_t * pes_v;

int first_chunk;

int skiptorecord;

/* Our counter for the number of chunks in our chunk buffer */
int realchunk_nr;

/* Our two main pes structurs where we store the data */
video_pes_t * m_video_pes;
audio_pes_t * m_audio_pes;

/*Counters for the number of PES's in out video and audio pes structure*/
int nr_video_pes;
int nr_audio_pes;

/* Flag if the video_pes wasn't finsished in the previous chunk */
int video_pes_not_finished;

/* Flag if the audio_pes wasn't finsished in the previous chunk */
int audio_pes_not_finished;

/* Tmp file nr */
int video_file_nr;
int audio_file_nr;

/* Total chunk number */
int chunk_nr;

/* Debug level */
int debug;

/* Video sync */
uint64_t med_ticks;
uint64_t last_tick;

/* Video start SEQ */
uint64_t video_seq_start;

/* Audio check */
uint16_t wrong_audio;

/* Audio start tick */
uint64_t audio_start;

/* Delay/Diff between start of Audio/Video */
int64_t es_delay;

/* Skipp Marker */
int vskipp;
int askipp;

/* Video Data Constants */
uint32_t pes_header_size;
int pes_time_offset;

/* Audio Data Constants */
int header_size;
uint32_t audio_pes_size;
int time_offset;

#ifdef A52
/* A52/AC3 */
int a52_okay;
#ifdef DEBUG
FILE * dbgfile;
#endif
#endif

/* Chunk size */
int chunk_size;

/* Write or not */
int write_file;

/* Start chunk */
int start_chunk;


/* End global */


int main(int argc, char *argv[]) {

	/* Args*/
	int flags;

	/* Switches */
	int audio_type;
	int jump_chunk;
	char * tmp_audio_type=NULL;
	char * tmp_debug=NULL;
	char * infile=NULL;
	char * audio_out=NULL;
	char * video_out=NULL;
	char * tmp_jump_chunk=NULL;


	/* Interation */
	int j;

	int f_chunk;

	uint64_t res_delay;

	/* Chunk related */
	chunk_t * chunk;
	chunk_t * chunks;

	/* Marker */
	int skipp;

	/* File handlers etc.*/
	int file_nr;
	off_t file_size;
	struct stat file_info;
	uint8_t tmp_v[]={0x00,0x00,0x01,0xe0};
	uint8_t tmp_ac3[]={0x00,0x00,0x01,0xbd};
	uint8_t tmp_mpeg[]={0x00,0x00,0x01,0xc0};

#ifdef A52
	/* A52/AC3 */
	static a52_state_t * state;
	uint32_t accel;
#endif



	/* Init global variables */

	/* Our two main pes structurs where we store the data */
	m_video_pes=NULL;
	m_audio_pes=NULL;

	/*Counters for the number of PES's in out video and audio pes structure*/
 	nr_video_pes=0;
	nr_audio_pes=0;

	/* Flag if the video_pes wasn't finsished in the previous chunk */
 	video_pes_not_finished = 0;

	/* Flag if the audio_pes wasn't finsished in the previous chunk */
 	audio_pes_not_finished = 0;

	/* Chunk size */
	chunk_size = 131072;

	/* Check them */
	first_chunk=1;

	skiptorecord = 0;
	/* End check them */

	/* Global audio constants
	SEE: probe_tystream*/

	/* Video constants */
	pes_header_size = 16;
	pes_time_offset = 9;

#ifdef A52
	/* A52/AC3 */
	a52_okay = 0;
#endif

	/* Video Sync */

	/* Setting max_tick to 4505 a average NTCS tick diff
	between frames/fields - This is only used for the first
	chunk */

	med_ticks=4505;

	/* Video start SEQ */
	video_seq_start = 0;

	/* Audio start tick */
	audio_start = 0;

	/* Delay/Diff between start of Audio/Video */
	es_delay=0;

	/* Skipp Marker */
	vskipp=0;
	askipp=0;

	/* Init the last video PES/PTS */
	last_tick = 0;


	/* Debug */
	debug = 1;

	/* Write or not */
	write_file=1;

	/* Start Chunk */
	start_chunk = 0;

	/* Video pes header find array */
	pes_v = (uint8_t *)malloc(sizeof(uint8_t) * 4);
	memcpy(pes_v, tmp_v, 4);

	/* End init global - note some are inited a bit later and in probe_tystream*/


	/* Command line parsing */

	while ((flags = getopt(argc, argv, "n?hi:v:a:p:t:d:j:")) != -1) {
		switch (flags) {
		case 'i':
			if(optarg[0]=='-') usage();
			infile = optarg;
			break;
		case 'a':
			if(optarg[0]=='-') usage();
			audio_out = optarg;
			break;
		case 'v':
			if(optarg[0]=='-') usage();
			video_out = optarg;
			break;
		case 'd':
			if(optarg[0]=='-') usage();
			tmp_debug = optarg;
			break;
		case 't':
			if(optarg[0]=='-') usage();
			tmp_audio_type = optarg;
			break;
		case 'j':
			if(optarg[0]=='-') usage();
			tmp_jump_chunk = optarg;
			break;
		case 'n':
			write_file = 0;
			break;
		case '?':
			usage();
			break;
		case 'h':
			usage();
			break;
		default:
			usage();
		}
	}


	if (infile == NULL) {
		usage();
		exit(1);
	}

	if(tmp_debug == NULL) {
		debug = 1;
	} else {
		debug = atoi(tmp_debug);
	}

	if(write != 0) {
		if (audio_out == NULL) {
			usage();
			exit(1);
		}
		if (video_out == NULL) {
			usage();
			exit(1);
		}
	}


	audio_type = 0;
	if(tmp_audio_type != NULL){
		audio_type = atoi(tmp_audio_type);
		if (audio_type != 1 && audio_type != 2){
			usage();
		}
	}


	if(tmp_jump_chunk != NULL && tmp_audio_type == NULL) {
		usage();
	}

	if(tmp_jump_chunk != NULL) {
		jump_chunk = atoi(tmp_jump_chunk);
	}

	 /* Open the TY file */
#ifdef WIN32
	file_nr = open(infile, O_RDONLY | _O_BINARY);
#else   /* Unix */
	file_nr = open(infile, O_RDONLY);
#endif

	if(file_nr == -1) {
		perror("");
		exit(-1);
	}

	fstat(file_nr, &file_info);
	file_size = file_info.st_size;



	/* Audio type */
	if(audio_type == 0){
		/* Probe for Audio type */
		if(probe_tystream(file_nr, file_size) == 0) {
			exit(-1);
		}
	} else if (audio_type == 1) {

		/* Set AC3 Audio */
		pes_a = (uint8_t *)malloc(sizeof(uint8_t) * 4);

		/* Set global audio constants */

		header_size=14;
		audio_pes_size = 1536;
		time_offset = 9;

		/* Audio Type */
		wrong_audio = 0x3c0;

		memcpy(pes_a, tmp_ac3, 4);

		/* END - Set global audio constants */
		if (tmp_jump_chunk != NULL) {
			lseek(file_nr, jump_chunk * chunk_size, SEEK_SET);
			start_chunk = jump_chunk;
		}

	} else if (audio_type == 2) {
		/* Set mpeg Audio */
		pes_a = (uint8_t *)malloc(sizeof(uint8_t) * 4);

		/* Set global audio constants */

		header_size=11;
		audio_pes_size = 480;
		time_offset = 6;

		/* Audio Type */
		wrong_audio = 0x9c0;

		memcpy(pes_a, tmp_mpeg, 4);


		/* END - Set global audio constants */
		if(tmp_jump_chunk != NULL) {
			lseek(file_nr, jump_chunk * chunk_size, SEEK_SET);
			start_chunk = jump_chunk;
			if(debug >= 1) {
				printf("Skipping to chunk %i - reseting chunk number\n", start_chunk);
			}
		}

	} else {
		if(debug >= 1) {
			printf("Audio Error: this is not supposed to happen\n");
		}
		exit(1);
	}

#ifdef A52
	if(header_size == 14) {
			a52_okay = 1;
			/* Init ffmpeg */
			avcodec_init();
			avcodec_register_all();
	}
#endif
#ifdef A52_2
	if(header_size == 14) {
		/* A52 init */
		state = a52_init (accel);
		if (state == NULL) {
			if(debug >= 1) {
				printf ("A52 init failed\n");
			}
			a52_okay = 0;
		} else {
			a52_okay = 1;
			/* Init ffmpeg */
			avcodec_init();
			avcodec_register_all();
		}
	}
#endif


	if(write_file) {
		/* Open output files */

#ifdef WIN32
		video_file_nr = open(video_out,O_WRONLY | O_CREAT | O_TRUNC | _O_BINARY, S_IREAD|S_IWRITE);
		audio_file_nr = open(audio_out,O_WRONLY | O_CREAT | O_TRUNC | _O_BINARY, S_IREAD|S_IWRITE);
#else
		video_file_nr = open(video_out,O_WRONLY | O_CREAT | O_TRUNC, S_IREAD|S_IWRITE);
		audio_file_nr = open(audio_out,O_WRONLY | O_CREAT | O_TRUNC, S_IREAD|S_IWRITE);
#endif


		if(video_file_nr == -1 || audio_file_nr == -1){
			perror("");
			exit(-1);
		}
#ifdef A52
#ifdef DEBUG
		dbgfile = fopen("dbgaudio.pcm", "w");
		if (!dbgfile) {
			perror("");
			exit(-1);
		}
#endif
#endif		

	}

	f_chunk = 0;
	skipp = 0;

	for(j=0, chunk_nr = 0, realchunk_nr = 0; j < file_size - (start_chunk * chunk_size); j = (j + chunk_size), chunk_nr++, realchunk_nr++) {

		if(debug >= 1 && debug < 3) {
			prog_meter(chunk_nr);
		}

		if(debug >= 3) {
			printf("Main: chunk %i realchunk: %i\n",chunk_nr, realchunk_nr);
		}

		chunk = read_chunk(file_nr);

		if(chunk == NULL ){
			realchunk_nr--;

		} else {
			chunk->next = NULL;
			chunk->previous = NULL;

			if(check_chunk(chunk) != NULL) {

				if(f_chunk) {
					chunks = add_chunk(chunk, chunks);
				} else {
					chunks = chunk;
					f_chunk = 1;
				}

			} else {
				realchunk_nr--;
				free_chunk(chunk);
			}

		}

		if(realchunk_nr > 2 && chunks != NULL) {
				chunks = parse_chunk(chunks);
		}

		if( video_seq_start != 0 && audio_start != 0 && skipp != 1) {
			if(video_seq_start > audio_start) {
				if(debug >= 3) {
					printf("Main: Audio starts %llu ticks ahead\n",video_seq_start - audio_start);
				}
				askipp = 1;
				skipp =1;
			} else if (video_seq_start < audio_start) {
				if(debug >= 3) {
					printf("Main: Video starts %llu ticks ahead\n",audio_start - video_seq_start);
				}
				vskipp = 1;
				skipp =1;
			} else {
				if(debug >= 3) {
					printf("Main: Video starts %llu ticks Audio starts at %llu ticks\n", video_seq_start, audio_start);
				}

				askipp = 0;
				vskipp = 0;
				skipp = 1;
				es_delay = 0;
			}
		}

		while(nr_video_pes > 20 && m_video_pes != NULL && skipp == 1) {
			m_video_pes = write_video_pes(m_video_pes);
		}
		while(nr_audio_pes > 20 && m_audio_pes != NULL && skipp == 1) {
			//printf("Writing audio_pes\n");
			m_audio_pes = write_audio_pes(m_audio_pes);
		}

		/**
		 * lseek to the next chunk there is sometimes
		 * garbage in the end of the chunk, beside
		 * Sometimes the last record ends like chunksize - XX bytes
		 * se we can't be sure that we are at the end of the chunk when
		 * we finished reading it.
		 */

		lseek(file_nr, (chunk_nr + 1) * 128*1024, SEEK_SET);


	}

	/* FIXME we loose 2 chunks plus 5 video and audio pes records */
	free_chunks(chunks);
	free_video_pess(m_video_pes);
	free_audio_pess(m_audio_pes);
	free(pes_a);
	free(pes_v);
#ifdef A52_2
	if(a52_okay) {
		a52_free(state);
	}
#endif
	if(debug >= 1) {

		if(es_delay >= 0) {
			res_delay = es_delay / 90;
			printf("\nAudio/Video Sync Offset: %lld ms (i.e. audio plays early) \n", res_delay);
		} else {
			printf("\nAudio/Video Sync Offset: %lld ms (i.e. audio plays late) \n", -1 * es_delay/90);
		}
	}
	return(0);
}


/* Misc functions */

uint64_t get_time(uint8_t * pt){

  uint64_t time;
  if(debug >= 7) {
  	printf("Get Time: %02x %02x %02x %02x %02x\n", pt[0], pt[1], pt[2], pt[3], pt[4]);
  }

  time = (((pt[0] & 0x0E) << 29) | (pt[1] << 22) | ((pt[2] & 0xFE) << 14) | (pt[3] << 7) | (pt[4] >> 1));
  return(time);

}


int probe_tystream(int file_nr, off_t file_size) {

	/* Interation */
	int j,i;

	/* Chunks */
	int lnr_chunks;
	int probe_hz;
	int chunk_lnr;

	chunk_t * chunk;


	/* Audio */
	int mpeg;
	int ac3;
	uint16_t right_audio;

	uint8_t tmp_ac3[]={0x00,0x00,0x01,0xbd};
	uint8_t tmp_mpeg[]={0x00,0x00,0x01,0xc0};


	pes_a = (uint8_t *)malloc(sizeof(uint8_t) * 4);

	ac3=0;
	mpeg=0;

	lnr_chunks = file_size/chunk_size;
	probe_hz = lnr_chunks/10;

	if(debug >= 1) {
		printf("Probing TyStream for audio type\n");
	}

	for(j=0, chunk_lnr=0;  j < file_size; j = (j + chunk_size*probe_hz), chunk_lnr = chunk_lnr + probe_hz ){

		chunk = read_chunk(file_nr);
		if(chunk != NULL) {
			for(i=0; i < chunk->nr_records; i++){
				switch(chunk->record_header[i].type) {
					case 0x3c0: /* Audio PES - MPEG */
						mpeg++;
						if(debug >= 5) {
							printf("Probe: Got mpeg - %i\n", mpeg);
						}
						break;
					case 0x9c0:
						ac3++;
						if(debug >= 5) {
							printf("Probe: Got ac3 - %i\n", ac3);
						}
						break;
				}
			}
			free_chunk(chunk);
		}
		lseek(file_nr, (chunk_lnr + probe_hz) * chunk_size, SEEK_SET);
	}

	/* Reset file */
	lseek(file_nr, 0, SEEK_SET);

	if(debug >= 5) {
		printf("Probe: Mpeg: %i AC3: %i\n", mpeg, ac3);
	}


	if(mpeg > ac3) {
		/* We got mpeg Audio */

		/* Set global audio constants */

		header_size=11;
		audio_pes_size = 480;
		time_offset = 6;

		/* Audio Type */
		wrong_audio = 0x9c0;
		right_audio = 0x3c0;

		memcpy(pes_a, tmp_mpeg, 4);

		/* END - Set global audio constants */
		if(debug >= 1) {
			printf("MPEG Layer II audio\n");
			printf("Setting audio frame size: %i\n",audio_pes_size);
		}
		get_start_chunk(file_nr, right_audio, file_size);
		return(2);

	} else if ( ac3 > mpeg) {
		/* We got AC3 Audio */

		/* Set global audio constants */

		header_size=14;
		audio_pes_size = 1536;
		time_offset = 9;

		/* Audio Type */
		wrong_audio = 0x3c0;
		right_audio = 0x9c0;

		memcpy(pes_a, tmp_ac3, 4);

		/* END - Set global audio constants */
		if(debug >= 1) {
			printf("AC3 (Dolby Digital) audio\n");
			printf("Setting audio frame size: %i\n",audio_pes_size);
		}
		get_start_chunk(file_nr, right_audio, file_size);
		return(1);

	} else {
		if(debug >= 1) {
			printf("Warning: Probe faild - specify audio type\n");
		}
		return(0);
	}
}



void get_start_chunk(int file_nr, uint16_t right_audio, off_t file_size) {

	/* interation */
	int j, i;

	/* Marker */
	int gotit;

	/* Chunks */
	int chunk_lnr;
	chunk_t * chunk;

	if(debug >= 1) {
		if(right_audio == 0x3c0) {
			printf("Seeking TyStream start of MPEG Layer II audio \n");
		} else {
			printf("Seeking TyStream start of AC3 (Dolby Digital) audio \n");
		}
	}



	gotit=0;


	for(j=0, chunk_lnr=0;  j < file_size; j = (j + chunk_size), chunk_lnr++){

		chunk = read_chunk(file_nr);
		if(chunk != NULL) {
			for(i=0; i < chunk->nr_records; i++){
				if(chunk->record_header[i].type == right_audio){
					gotit=1;
					break;
				}
			}
			free_chunk(chunk);
		}
		if(gotit){
			break;
		}
		lseek(file_nr, (chunk_lnr + 1) * chunk_size, SEEK_SET);
	}
	lseek(file_nr, chunk_lnr * chunk_size, SEEK_SET);
	start_chunk = chunk_lnr;

	if(debug >= 1) {
		if(right_audio == 0x3c0) {
			printf("Found start of MPEG Layer II audio \n");
		} else {
			printf("Found start of AC3 (Dolby Digital) audio \n");
		}
		printf("Skipping to chunk %i - reseting chunk numbering \n", start_chunk);
	}

}

void prog_meter(int chunk_lnr) {

	if( chunk_lnr < 1000) {
		if(chunk_lnr%10 == 0 && chunk_lnr != 0 && chunk_lnr%500 != 0 && chunk_lnr%100 != 0) {
			printf(".");
		}

		if(chunk_lnr%100 == 0 && chunk_lnr != 0 && chunk_lnr%500 != 0) {
			printf("  %i", chunk_lnr);
		}

		if(chunk_lnr%500 == 0 && chunk_lnr != 0 ) {
			printf("  %i\n",chunk_lnr);
		}
	} else if (chunk_lnr >= 1000 && chunk_lnr < 10000) {
		if(chunk_lnr%10 == 0 && chunk_lnr != 0 && chunk_lnr%500 != 0 && chunk_lnr%100 != 0) {
			printf(".");
		}

		if(chunk_lnr%100 == 0 && chunk_lnr != 0 && chunk_lnr%500 != 0) {
			printf(" %i", chunk_lnr);
		}

		if(chunk_lnr%500 == 0 && chunk_lnr != 0 ) {
			printf(" %i\n",chunk_lnr);
		}

	} else {
		if(chunk_lnr%10 == 0 && chunk_lnr != 0 && chunk_lnr%500 != 0 && chunk_lnr%100 != 0) {
			printf(".");
		}

		if(chunk_lnr%100 == 0 && chunk_lnr != 0 && chunk_lnr%500 != 0) {
			printf("%i", chunk_lnr);
		}

		if(chunk_lnr%500 == 0 && chunk_lnr != 0 ) {
			printf("%i\n",chunk_lnr);
		}

	}
	fflush(stdout);

}

void usage(){

	printf("\ntydemux %i.%i.%i usage:\n"
	       "tydemux [option] -i infile -a outfile_audio -v outfile_video\n"
	       "\n"
	       "Where option is:\n"
	       "\t -d 0-9\t(debug level 0 == silent, 9 == very verbose, default == 1)\n"
	       "\t -t 1/2\t(set audio type to 1 == ac3 or 2 == mpeg - disables audio probe)\n"
	       "\t -j n\t(set start chunk to n i.e. jump to chunk n, n starts at zero)\n"
	       "\t -n\t(no write - dry run don't need output files)\n"
	       "\t -h/?\t(help - writes this message)\n\n", MAJOR_VER, MINOR_VER,SUB_VER);
	exit(1);
}
















