/*
 * Copyright (C) 2002  Olaf Beck <olaf_sc@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "tydemux.h"
#include "global.h"

void free_video_pes(video_pes_t * video_pes) {
	free(video_pes->pes_header);
	if( video_pes->pes_data != NULL) {
		free(video_pes->pes_data);
	}
	free(video_pes);
}

void free_video_pess(video_pes_t * video_pes) {

	video_pes_t * free_pes;

	while(video_pes->next){
		free_pes = video_pes;
		video_pes = video_pes->next;
		free_video_pes(free_pes);
	}

	free_video_pes(video_pes);
}


video_pes_t * write_video_pes(video_pes_t * video_pes) {

	video_pes_t * next_pes;
	nr_video_pes--;

	if(debug >= 3) {
		printf("Write Video: writing video pes: time: %llu  size: %d\n",video_pes->time,video_pes->size);
	}

	if(video_pes->next != NULL) {
		/* Writing :) */
		if(!vskipp) {
			if(write_file){
				//write(video_file_nr, video_pes->pes_header, 16);
				write(video_file_nr, video_pes->pes_data, video_pes->size);
			}
			next_pes=video_pes->next;
			next_pes->previous = NULL;
			free_video_pes(video_pes);
			return(next_pes);
		} else {
			if(video_pes->time <= audio_start && video_pes->next->time >= audio_start ) {
				/* We got a match
				Lets see which one is the closest*/
				if(debug >= 3) {
					printf("Write Video: Delay match - time %llu - next time %llu - audio %llu\n", video_pes->time,video_pes->next->time,audio_start);
				}
				if((audio_start - video_pes->time) <= (video_pes->next->time - audio_start)) {
					/* Okay we should write this one and set the delay to XXX ticks
					early start of video stream i.e negative value*/
					if(write_file){
						//write(video_file_nr, video_pes->pes_header, 16);
						write(video_file_nr, video_pes->pes_data, video_pes->size);
					}
					es_delay = video_pes->time - audio_start;
					vskipp = 0;
					next_pes=video_pes->next;
					next_pes->previous = NULL;
					free_video_pes(video_pes);
					return(next_pes);
				} else {
					/* Jump to next pes and write it set the delay to XXX ticks
					late start of video stream i.e positive value*/
					next_pes=video_pes->next;
					next_pes->previous = NULL;
					free_video_pes(video_pes);
					video_pes = next_pes;

					if(write_file){
						//write(video_file_nr, video_pes->pes_header, 16);
						write(video_file_nr, video_pes->pes_data, video_pes->size);
					}

					es_delay = video_pes->time - audio_start;
					vskipp = 0;

					/* FIXME we should check the next pes isn't NULL */
					next_pes=video_pes->next;
					next_pes->previous = NULL;
					free_video_pes(video_pes);
					return(next_pes);
				}

			} else {
				/* bummer not this one either :) */
				next_pes=video_pes->next;
				next_pes->previous = NULL;
				free_video_pes(video_pes);
				return(next_pes);
			}
		}
 	} else {
		/* last pes to write - we don't do any skipp search since this is higly unlikely */
		if(write_file){
			//write(video_file_nr, video_pes->pes_header, 16);
			write(video_file_nr, video_pes->pes_data, video_pes->size);
		}
		free(video_pes);
		return(0);
	}
}




video_pes_t * insert_video_pes(video_pes_t * video_pes){


	video_pes_t * tmp_video_pes;

	/* Increase the video pes counter */
	nr_video_pes++;

	tmp_video_pes = m_video_pes;

	while (tmp_video_pes->next) {
		tmp_video_pes = tmp_video_pes->next;
	}
	tmp_video_pes->next = video_pes;
	video_pes->previous = tmp_video_pes;
	video_pes->next = NULL;
	return(m_video_pes);

}




video_pes_t * get_last_video_pes(){


	video_pes_t * tmp_video_pes;
	video_pes_t * video_pes;

	/* decrease the video pes counter */
	nr_video_pes--;

	tmp_video_pes = m_video_pes;

	while (tmp_video_pes->next) {
		tmp_video_pes = tmp_video_pes->next;
	}

	tmp_video_pes = tmp_video_pes->previous;
	video_pes = tmp_video_pes->next;
	tmp_video_pes->next = NULL;
	video_pes->previous = NULL;
	return(video_pes);
}

int find_seq_start(int i, chunk_t * chunks, int size, uint8_t * tmp_video_buffer) {

	int k;
	int tsize;
	int gotit;
	int gotoffset;
	int offset;
	uint8_t * pt1;
	uint8_t seq_header_start[]={0x00, 0x00, 0x01, 0xb3};

	gotit = 0;

	for(;i < chunks->nr_records; i++) {
		switch(chunks->record_header[i].type){
			case 0x7e0:
				tsize = size;
				size = size + chunks->record_header[i].size;
				tmp_video_buffer = (uint8_t *)realloc(tmp_video_buffer, size * sizeof(uint8_t));
				pt1 = tmp_video_buffer + tsize;
				memcpy(pt1, chunks->records[i].record,
					sizeof(uint8_t) * chunks->record_header[i].size);
				break;
			case 0x2e0:
			case 0x8e0:
			case 0xae0:
			case 0xbe0:
			case 0xce0:
			case 0x6e0:
				tsize = size;
				size = size + chunks->record_header[i].size;
				tmp_video_buffer = (uint8_t *)realloc(tmp_video_buffer, size * sizeof(uint8_t));
				pt1 = tmp_video_buffer + tsize;
				memcpy(pt1, chunks->records[i].record,
					sizeof(uint8_t) * chunks->record_header[i].size);
				gotit = 1;
				break;
		}
		if(gotit == 1){
			break;
		}

	}

	gotoffset=0;
	if(gotit) {
		pt1 = tmp_video_buffer;
		/* FIXME need to check if we should habe header_size or (header_size + 1) since the size of the header is header_size */
		for(k=0; k < size - (4 + 1); k++, pt1++) {
			if (memcmp(pt1, seq_header_start, 4) == 0) {
				offset = pt1 - tmp_video_buffer;
				free(tmp_video_buffer);
				gotoffset=1;

			}
			if(gotoffset) {
				break;
			}
		}
	}

	if(gotoffset) {
		return(offset);
	} else if (chunks->next != NULL && gotit != 1) {
		return(find_seq_start(0, chunks->next, size, tmp_video_buffer));
	} else {
		printf("Find SEQ Start: ERROR!! - can't find start of seq header - exiting\n");
		exit(-1);
	}

}


void parse_video(chunk_t * chunks){

	/* Interation */
	int i, k;
	uint32_t p;

	/* Markers */
	int gotit;

	/*Size in the chunk*/
	int size;
	int tsize;


	/* Video PES */
	video_pes_t * video_pes;
	uint8_t * pes_header;
	uint8_t * pes_data=NULL;

	/* Pointers to magic with */
	uint8_t * pt1;

	/* SEQ buffer */
	uint8_t * seq_buffer;
	int offset;


	if(debug >= 3) {
		printf("Parse Video: chunk %i\n", chunk_nr - 2);
	}


	i=0;

	if(video_pes_not_finished){

		if(debug >= 6) {
			printf("Parse Video: overlap from chunk %i \n", chunk_nr - 1);
		}

		video_pes = get_last_video_pes();

		gotit = 0;
		size = video_pes->size;
		pes_data=video_pes->pes_data;

		for(; i < chunks->nr_records; i++) {
			/* Okay copy all Video records of type
			0x2e0, 0x7e0, 0x8e0, 0xae0, 0xbe0, 0xce0 into the pes
			The rest is either audio or junk (CC,TT EE type records)

			We are doing this until we find the next video pes or
			the end of the chunk. If we find the end of the chunk
			we flag the video_pes_not_finished unless it's the last
			chunk in our file. If so we drop the pes */

			switch(chunks->record_header[i].type){
				case 0x2e0:
				case 0x7e0:
				case 0x8e0:
				case 0xae0:
				case 0xbe0:
				case 0xce0:
					tsize = size;
					size = size + chunks->record_header[i].size;
					video_pes->size = size;
					pes_data = (uint8_t *)realloc(pes_data, size * sizeof(uint8_t));
					pt1 = pes_data + tsize;
					memcpy(pt1, chunks->records[i].record,
						sizeof(uint8_t) * chunks->record_header[i].size);
					video_pes->pes_data = pes_data;
					break;
				case 0x6e0:
					gotit=1;
					break;
			}
			if(gotit == 1){
				video_pes_not_finished = 0;
				break;
			}

		}

		/* FIXME if gotis is 0 and this is the last chunk
		we should not insert this pes !*/


		if(gotit == 0) {
			/* This mean that we reach the end of the chunk but
			didn't find the end of our present video pes. We now rise
			the video_pes_not_finished flag unless it's the last chunk */
			if(chunks->next != NULL) {
				video_pes_not_finished = 1;
				if(debug >= 3) {
					printf("Parse Video: pes video overlap - video_pes_not_finished\n");
				}
				if(m_video_pes != NULL){
					m_video_pes = insert_video_pes(video_pes);
				} else {
					m_video_pes = video_pes;
				}
				/* Well since there is no more in this chunk just return */
				return;
			}
		} else {
			if(m_video_pes != NULL){
				m_video_pes = insert_video_pes(video_pes);
			} else {
				m_video_pes = video_pes;
			}
		}

	}


	for(;i < chunks->nr_records; i++) {

		if(debug >= 6) {
			printf("Parse Video: record %i of %i records - chunk %i\n",i + 1, chunks->nr_records, chunk_nr -2 );
		}

		switch(chunks->record_header[i].type){
			case 0x6e0:
				if(debug >= 6){
					printf("Parse Video: got pes header - record %i \n", i + 1);
				}

				/* Allocate and init the pes structure */
				video_pes = (video_pes_t *)malloc(sizeof(video_pes_t));
				video_pes->pes_data=NULL;
				video_pes->pes_header=NULL;
				video_pes->next=NULL;
				video_pes->previous=NULL;
				video_pes->time = 0;
				video_pes->size = 0;

				/* Find PTS time of PES */
				pt1 = chunks->records[i].record;
				pt1 = pt1 + pes_time_offset;
				video_pes->time = get_time(pt1);

				/* Copy pes header to pes structure
				The video header is "always" 16 bytes
				FIXME need to be checked */
				pes_header = (uint8_t *)malloc(sizeof(uint8_t) * pes_header_size);
				memcpy(pes_header,chunks->records[i].record,sizeof(uint8_t) * pes_header_size);
				video_pes->pes_header = pes_header;
				if(debug >= 1) {
					if(chunks->record_header[i].size > pes_header_size) {
							printf("Parse Video: pes video header size error: %i - chunk %i\n", chunks->record_header[i].size, chunk_nr -2);
					}
				}

				/* Find end of PES since DTV is useing a TS stream the pes
				header is lacking the size field of the pes payload :(
				Well it's zero which is leagal in a Video stream of TS stream
				see: 2.4.3.7 in the Mpeg System std*/

				gotit=0;
				size=0;
				pes_data=NULL;
				for(i=i+1; i < chunks->nr_records; i++) {
					/* Okay copy all Video records of type
					0x2e0, 0x7e0, 0x8e0, 0xae0, 0xbe0, 0xce0 into the pes
					The rest is either audio or junk (CC,TT EE type records)

					We are doing this until we find the next video pes or
					the end of the chunk. If we find the end of the chunk
					we flag the video_pes_not_finished unless it's the last
					chunk in our file. If so we drop the pes */

					switch(chunks->record_header[i].type){

						case 0x7e0:
							if(video_seq_start == 0) {
								seq_buffer = NULL;
								offset = find_seq_start(i, chunks, 0, seq_buffer);
								video_seq_start = video_pes->time;
								if(debug >= 1) {
									printf("Parse Video: Start SEQ header at %llu ticks\n",video_seq_start);
								}
							} else {
								offset = 0;

							}
							tsize = size;
							size = size + chunks->record_header[i].size - offset;
							video_pes->size = size;
							pes_data = (uint8_t *)realloc(pes_data, size * sizeof(uint8_t));
							pt1 = pes_data + tsize;
							memcpy(pt1, chunks->records[i].record + offset,
								(sizeof(uint8_t) * chunks->record_header[i].size) - offset);
							video_pes->pes_data = pes_data;
							break;
						case 0x2e0:
						case 0x8e0:
						case 0xae0:
						case 0xbe0:
						case 0xce0:
							if ( video_seq_start != 0) {
								tsize = size;
								size = size + chunks->record_header[i].size;
								video_pes->size = size;
								pes_data = (uint8_t *)realloc(pes_data, size * sizeof(uint8_t));
								pt1 = pes_data + tsize;
								memcpy(pt1, chunks->records[i].record,
									sizeof(uint8_t) * chunks->record_header[i].size);
								video_pes->pes_data = pes_data;
							}
							break;
						case 0x6e0:
							gotit=1;
							break;
					}
					if(gotit == 1){
						/* Reset i so the catch the next pes */
						i--;
						break;
					}

				}

				if(debug >= 6) {
					printf("Parse Video: Header:");
					for (k = 0; k < 16; k++){
						printf(" %02x", video_pes->pes_header[k]);
					}
					printf("\nParse Video: Data:");
					for (p = 0; p < 16 && p < video_pes->size; p++){
						printf(" %02x", video_pes->pes_data[p]);
					}
					printf("\n");
				}

				if(gotit == 0) {
					/* This mean that we reach the end of the chunk but
					didn't find the end of our present video pes. We now rise
					the video_pes_not_finished flag unless it's the last chunk */
					if(chunks->next != NULL) {

						if(debug >= 6) {
							printf("Parse Video: pes video - video_pes_not_finished\n");
						}
						if (video_seq_start != 0) {
							if(m_video_pes != NULL){
								m_video_pes = insert_video_pes(video_pes);
							} else {
								m_video_pes = video_pes;
							}
							/* Well since there is no more in this chunk just return */
							video_pes_not_finished = 1;
						} else {
							free_video_pes(video_pes);
							video_pes_not_finished = 0;
						}
						return;
					}
				} else {
					if (video_seq_start != 0) {
						if(m_video_pes != NULL){
							m_video_pes = insert_video_pes(video_pes);
						} else {
							m_video_pes = video_pes;
						}
					} else {
						free_video_pes(video_pes);
						video_pes_not_finished = 0;
					}
				}
				break;
		}
	}

}
